'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.List = undefined;

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }(); /**
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      * @author Grzegorz Rozdzialik <voreny.gelio@gmail.com>
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      */


var _node = require('./node');

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

/**
 * Javascript list implementation
 *
 * @export
 * @class List
 */
var List = exports.List = function () {
  /**
   * Creates an instance of List.
   *
   * @param {any} initialValue  Value to initiate the list with.
   */
  function List(initialValue) {
    _classCallCheck(this, List);

    if (initialValue) {
      this.head = new _node.Node(initialValue);
    }
  }

  /**
   * Computes the length of the list
   *
   * @return {number} The length of the list.
   */


  _createClass(List, [{
    key: 'getLength',
    value: function getLength() {
      var current = this.head;
      var length = 0;

      while (current) {
        current = current.next;
        length += 1;
      }

      return length;
    }

    /**
     * Returns the head of the list (the first element)
     *
     * @return {Node} The head node (the first node in the list)
     */

  }, {
    key: 'getHead',
    value: function getHead() {
      return this.head;
    }

    /**
     * Returns the tail of the list (the last element)
     *
     * @return {Node} The tail node (the last node in the list)
     */

  }, {
    key: 'getTail',
    value: function getTail() {
      return this.reduce(function (_, node) {
        return node;
      }, undefined, false);
    }

    /**
     * @callback reduceCallbackFn
     * @param {any} accumulated
     * @param {any} current
     * @return {any}
     */

    /**
     * Reduces the list to a single value
     *
     * @param {reduceCallbackFN} callbackFn   Callback which reduces the list.
     * @param {any} [startingValue]           Value to initiate the reducing with.
     * @param {boolean} [extractValues=true]  Decides on what will be passed to the callbackFn,
     * either values or whole nodes.
     * @return {any}                          Reduced value
     */

  }, {
    key: 'reduce',
    value: function reduce(callbackFn, startingValue) {
      var extractValues = arguments.length <= 2 || arguments[2] === undefined ? true : arguments[2];

      var currentNode = void 0;
      var accumulated = void 0;

      var extractorFn = void 0;
      if (extractValues) {
        extractorFn = function extractorFn(node) {
          return node.value;
        };
      } else {
        extractorFn = function extractorFn(node) {
          return node;
        };
      }

      if (!this.head) {
        return startingValue;
      }

      if (startingValue === undefined) {
        currentNode = this.head;
        accumulated = startingValue;
      } else {
        currentNode = this.head.next;
        accumulated = extractorFn(this.head);
      }

      while (currentNode) {
        accumulated = callbackFn(accumulated, extractorFn(currentNode));
        currentNode = currentNode.next;
      }

      return accumulated;
    }

    /**
     * @callback valueCallbackFn
     * @param {any} value
     * @param {number} index
     */

    /**
     * Traverses the list and executes the callback function for each element in the list.
     * The first argument of the callback function is the node's value, the second one is the index.
     *
     * @param {valueCallbackFn} callbackFn  Function invoked for each element in the list.
     */

  }, {
    key: 'forEach',
    value: function forEach(callbackFn) {
      var index = 0;

      this.reduce(function (_, value) {
        callbackFn(value, index);
        index += 1;
        return null;
      });
    }

    /**
     * Executes callbackFn on each item from the list and returns the results as another list.
     *
     * @param {valueCallbackFn} callbackFn  Function invoked for each element in the list
     * @return {List}   Transformed list
     */

  }, {
    key: 'map',
    value: function map(callbackFn) {
      var outputList = new List();

      this.forEach(function (value, index) {
        var transformedValue = callbackFn(value, index);
        outputList.pushBack(transformedValue);
      });

      return outputList;
    }

    /**
     * Retrieves the Node with a given index
     *
     * @param {number} targetIndex  Wanted node's index
     * @return {Node}               Found node
     */

  }, {
    key: 'get',
    value: function get(targetIndex) {
      var currentIndex = 0;
      var currentNode = this.head;

      while (currentIndex < targetIndex && currentNode) {
        currentIndex += 1;
        currentNode = currentNode.next;
      }

      return currentNode;
    }

    /**
     * Adds a value to the end of the list
     *
     * @param {any} newValue  A value to be added
     * @return {List}         This list. Allows for chainability
     */

  }, {
    key: 'pushBack',
    value: function pushBack(newValue) {
      var newNode = new _node.Node(newValue);
      var tail = this.getTail();

      if (tail) {
        tail.next = newNode;
      } else {
        this.head = newNode;
      }

      return this;
    }

    /**
     * Adds a value at a given index
     *
     * @param {any} value     Value to be added.
     * @param {number} index  Index at which the value should be added.
     * @throws {Error} Index must not exceed list size.
     * @return {List}    This list. Allows for chainability.
     */

  }, {
    key: 'push',
    value: function push(value, index) {
      if (index === 0) {
        // Replace head
        this.head = new _node.Node(value, this.head);
      } else {
        var previousNode = this.get(index - 1);
        if (!previousNode) {
          throw new Error('Index exceeds list\'s size');
        }

        var nextNode = previousNode.next;
        previousNode.next = new _node.Node(value, nextNode);
      }

      return this;
    }

    /**
     * Removes nodes with specific values from the list
     *
     * @param {any} valueToRemove A value to be removed
     * @return {List}             This list. Allows for chainability.
     */

  }, {
    key: 'remove',
    value: function remove(valueToRemove) {
      if (this.head.value === valueToRemove) {
        this.head = this.head.next;
        return this.remove(valueToRemove);
      }

      this._removeRecursive(this.head, valueToRemove);
      return this;
    }

    /**
     * Used internally by remove. Replaces the current node with a next one if the value matches.
     *
     * @private
     * @param {Node} previousNode Reference to the previous node.
     * @param {any} valueToRemove Value to be removed.
     */

  }, {
    key: '_removeRecursive',
    value: function _removeRecursive(previousNode, valueToRemove) {
      if (!previousNode.next) {
        return;
      }
      var currentNode = previousNode.next;

      if (currentNode.value === valueToRemove) {
        previousNode.next = currentNode.next;
        this._removeRecursive(previousNode, valueToRemove);
      } else {
        this._removeRecursive(currentNode, valueToRemove);
      }
    }

    /**
     * Returns the first node that value matches.
     *
     * @param {any} value Value to be found
     * @return {Node}     Node with that value
     */

  }, {
    key: 'find',
    value: function find(value) {
      function findRecursive(node, value) {
        if (!node) {
          return null;
        } else if (node.value === value) {
          return node;
        }
        return findRecursive(node.next, value);
      }

      return findRecursive(this.head, value);
    }

    /**
     * Converts the list to an array
     *
     * @return {any[]}  Array of values from the list
     */

  }, {
    key: 'getValues',
    value: function getValues() {
      var valuesArray = [];
      this.forEach(function (value) {
        return valuesArray.push(value);
      });
      return valuesArray;
    }
  }]);

  return List;
}();